﻿/*
Copyright 2012, 2013 Adam Carter (http://adam-carter.com)

This file is part of FileCache (http://fc.codeplex.com).

FileCache is distributed under the Microsoft Public License (Ms-PL).
Consult "LICENSE.txt" included in this package for the complete Ms-PL license.
*/
using System.Runtime.Caching;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Runtime.Serialization;
using System.Collections.Generic;
using System.Threading;
using System.IO;

namespace FC.UnitTests
{


    /// <summary>
    ///This is a test class for FileCacheTest and is intended
    ///to contain all FileCacheTest Unit Tests
    ///</summary>
    [TestClass()]
    public class FileCacheTest
    {


        private TestContext testContextInstance;

        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext
        {
            get
            {
                return testContextInstance;
            }
            set
            {
                testContextInstance = value;
            }
        }

        #region Additional test attributes
        // 
        //You can use the following additional attributes as you write your tests:
        //
        //Use ClassInitialize to run code before running the first test in the class
        //[ClassInitialize()]
        //public static void MyClassInitialize(TestContext testContext)
        //{
        //}
        //
        //Use ClassCleanup to run code after all tests in a class have run
        //[ClassCleanup()]
        //public static void MyClassCleanup()
        //{
        //}
        //
        //Use TestInitialize to run code before running each test
        //[TestInitialize()]
        //public void MyTestInitialize()
        //{
        //}
        //
        //Use TestCleanup to run code after each test has run
        //[TestCleanup()]
        //public void MyTestCleanup()
        //{
        //}
        //
        #endregion

        [TestMethod]
        public void AbsoluteExpirationTest()
        {
            FileCache target = new FileCache();
            CacheItemPolicy policy = new CacheItemPolicy();

            //add an item and have it expire yesterday
            policy.AbsoluteExpiration = (DateTimeOffset)DateTime.Now.AddDays(-1);
            target.Set("test", "test", policy);

            //then try to access the item
            object result = target.Get("test");
            Assert.AreEqual(null, result);
        }

        [TestMethod]
        public void PolicySaveTest()
        {
            FileCache target = new FileCache();
            CacheItemPolicy policy = new CacheItemPolicy();
            policy.SlidingExpiration = new TimeSpan(1, 0, 0, 0, 0);
            target.Set("test", "test", policy);

            CacheItemPolicy returnPolicy = target.GetPolicy("test");
            Assert.AreEqual(policy.SlidingExpiration, returnPolicy.SlidingExpiration);
        }

        [TestMethod]
        public void SlidingExpirationTest()
        {
            FileCache target = new FileCache();
            CacheItemPolicy policy = new CacheItemPolicy();

            //add an item and have it expire 500 ms from now
            policy.SlidingExpiration = new TimeSpan(0, 0, 0, 0, 500);
            target.Set("test", "test", policy);

            //sleep for 200
            Thread.Sleep(200);

            //then try to access the item
            object result = target.Get("test");
            Assert.AreEqual("test", result);

            //sleep for another 200
            Thread.Sleep(200);

            //then try to access the item
            result = target.Get("test");
            Assert.AreEqual("test", result);

            //then sleep for more than 500 ms.  Should be gone
            Thread.Sleep(600);
            result = target.Get("test");
            Assert.AreEqual(null, result);
        }

        [TestMethod]
        public void CustomObjectSaveTest()
        {
            FileCache target = new FileCache();

            //create custom object
            CustomObjB custom = new CustomObjB()
            {
                Num = 5,
                obj = new CustomObjA()
                {
                    Name = "test"
                }
            };

            CacheItem item = new CacheItem("foo")
            {
                Value = custom,
                RegionName = "foobar"
            };

            //set it
            target.Set(item, new CacheItemPolicy());

            //now get it back
            CacheItem fromCache = target.GetCacheItem("foo", "foobar");

            //pulling twice increases code coverage
            fromCache = target.GetCacheItem("foo", "foobar");
            custom = null;
            custom = fromCache.Value as CustomObjB;
            Assert.IsNotNull(custom);
            Assert.IsNotNull(custom.obj);
            Assert.AreEqual(custom.Num, 5);
            Assert.AreEqual(custom.obj.Name, "test");
        }

        [TestMethod]
        public void RemoveTest()
        {
            FileCache target = new FileCache();
            target.Set("test", "test", DateTimeOffset.Now.AddDays(3));
            object result = target.Get("test");
            Assert.AreEqual("test", result);

            //now delete
            target.Remove("test");
            result = target["test"];
            Assert.AreEqual(null, result);
        }

        [TestMethod]
        public void TestCount()
        {
            FileCache target = new FileCache("testCount");
            target["test"] = "test";

            object result = target.Get("test");
            Assert.AreEqual("test", result);
            Assert.AreEqual(1, target.GetCount());
        }

        [TestMethod]
        public void DefaultRegionTest()
        {
            FileCache cacheWithDefaultRegion = new FileCache();
            cacheWithDefaultRegion.DefaultRegion = "foo";
            FileCache defaultCache = new FileCache();
            cacheWithDefaultRegion["foo"] = "bar";
            object pull = defaultCache.Get("foo", "foo");
            Assert.AreEqual("bar", pull.ToString());
        }

        [TestMethod]
        public void AccessTimeoutTest()
        {
            //AC: This test passes in debug mode, but not in rutime mode.  Why?

            FileCache target = new FileCache();
            target.AccessTimeout = new TimeSpan(1);
            target["primer"] = 0;
            string filePath = Path.Combine(target.CacheDir, "cache", "foo.dat");
            FileStream stream = File.Open(filePath, FileMode.Create);
            try
            {
                object result = target["foo"];

                //file access should fail.  If it doesn't, the test fails.
                Assert.AreNotEqual(true, true);
            }
            catch (IOException)
            {
                //we expect a file exception.
                Assert.AreEqual(true, true);
            }
            stream.Close();
        }

        [TestMethod]
        public void DefaultPolicyTest()
        {
            FileCache target = new FileCache();
            CacheItemPolicy policy = new CacheItemPolicy();
            policy.SlidingExpiration = new TimeSpan(1);
            target.DefaultPolicy = policy;
            target["foo"] = "bar";
            Thread.Sleep(2);
            object result = target["foo"];
            Assert.AreEqual(null, result);
        }

        [TestMethod]
        public void GetEnumeratorTest()
        {
            FileCache target = new FileCache();
            target["foo"] = 1;
            target["bar"] = 2;

            foreach (KeyValuePair<string, object> kvp in target)
            {
                Assert.AreEqual(target[kvp.Key], kvp.Value);
            }
        }
    }
}
